import os
import pathlib
import sys

import pyarrow
from pybind11.setup_helpers import Pybind11Extension, build_ext

__version__ = "0.2.0rc4"

ROOT = pathlib.Path(__file__).parent
README = (ROOT / "README.md").read_text()

USE_CXX11_ABI = os.environ.get("USE_CXX11_ABI", "0")


def build(setup_kwargs):

    try:
        pyarrow.create_library_symlinks()
    except FileExistsError:
        pass  # For some reason this started complaininig
    source_directory = ROOT / "cpp" / "src"
    extension = Pybind11Extension(
        name="_cassarrow",
        sources=[
            str(source_directory / "cassarrow/bindings.cpp"),
            str(source_directory / "cassarrow/cassarrow.cpp"),
        ],
        define_macros=[("VERSION_INFO", __version__)],
        cxx_std=17,
        library_dirs=pyarrow.get_library_dirs(),
        libraries=pyarrow.get_libraries(),
        runtime_library_dirs=[]
        if sys.platform == "win32"
        else pyarrow.get_library_dirs(),
        include_dirs=[source_directory, pyarrow.get_include()],
        extra_compile_args=[f"-D_GLIBCXX_USE_CXX11_ABI={USE_CXX11_ABI}"],
    )
    setup_kwargs.update(
        {
            "ext_modules": [extension],
            "cmd_class": {"build_ext": build_ext},
            "zip_safe": False,
        }
    )



import contextlib
import io
from typing import Iterator, List, Tuple

import cassandra.cluster
import pyarrow as pa  # Must be imported before _cassarrow
from cassandra.cqltypes import CassandraTypeType
from cassandra.protocol import ResultMessage, _ProtocolHandler

import _cassarrow

NATIVE_TYPES = {
    "ascii": pa.string(),
    "bigint": pa.int64(),
    "blob": pa.binary(),
    "boolean": pa.bool_(),
    "counter": pa.int64(),
    "date": pa.date32(),
    "double": pa.float64(),
    "duration": pa.duration("ns"),
    "float": pa.float32(),
    # "inet":
    "int": pa.int32(),
    "smallint": pa.int16(),
    "text": pa.string(),
    "time": pa.time64("ns"),
    "timestamp": pa.timestamp("ms"),
    "timeuuid": pa.binary(16),
    "tinyint": pa.int8(),
    "uuid": pa.binary(16),
    "varchar": pa.string(),
    # "varint":
}


def get_arrow_type(dtype: CassandraTypeType) -> pa.DataType:
    typename = dtype.typename
    cassname = dtype.cassname
    if cassname in ("ListType", "SetType"):
        assert len(dtype.subtypes) == 1
        return pa.list_(get_arrow_type(dtype.subtypes[0]))
    elif cassname == "MapType":
        assert len(dtype.subtypes) == 2
        return pa.map_(
            get_arrow_type(dtype.subtypes[0]),
            get_arrow_type(dtype.subtypes[1]),
            keys_sorted=True,
        )
    elif cassname == "UserType":
        return pa.struct(
            [
                pa.field(name, get_arrow_type(subtype))
                for name, subtype in zip(dtype.fieldnames, dtype.subtypes)
            ]
        )
    try:
        return NATIVE_TYPES[typename]
    except KeyError:
        raise TypeError(f"{typename}: {dtype}")


def column_metadata_to_schema(column_metadata: List[tuple]) -> pa.Schema:
    return pa.schema(
        [
            pa.field(column_name, get_arrow_type(dtype))
            for keyspace, table, column_name, dtype in column_metadata
        ]
    )


def metadata_to_schema(names: List[str], dtypes: List[CassandraTypeType]):
    return pa.schema(
        [pa.field(name, get_arrow_type(dtype)) for name, dtype in zip(names, dtypes)]
    )


class ArrowResultMessage(ResultMessage):
    code_to_type = dict((v, k) for k, v in ResultMessage.type_codes.items())

    def recv_results_rows(
        self,
        f: io.BytesIO,
        protocol_version: int,
        user_type_map: dict,
        result_metadata: list,
    ) -> None:
        assert protocol_version >= 3
        self.recv_results_metadata(f, user_type_map)
        column_metadata = self.column_metadata or result_metadata
        self.column_names = [c[2] for c in column_metadata]
        self.column_types = [c[3] for c in column_metadata]
        schema = column_metadata_to_schema(column_metadata)
        self.parsed_rows = _cassarrow.parse_results(f.read(), schema)


def result_set_to_table(result_set: cassandra.cluster.ResultSet) -> pa.Table:
    schema = metadata_to_schema(result_set.column_names, result_set.column_types)
    return pa.Table.from_batches(result_set, schema=schema)


class ArrowProtocolHandler(_ProtocolHandler):
    message_types_by_opcode = {
        **_ProtocolHandler.message_types_by_opcode,
        **{ArrowResultMessage.opcode: ArrowResultMessage},
    }


def record_batch_factory(
    colnames: List[str], rows: pa.RecordBatch
) -> Tuple[pa.RecordBatch]:
    return (rows,)


@contextlib.contextmanager
def install_cassarrow(
    session: cassandra.cluster.Session,
) -> Iterator[cassandra.cluster.Session]:
    row_factory, client_protocol_handler = (
        session.row_factory,
        session.client_protocol_handler,
    )
    session.row_factory, session.client_protocol_handler = (
        record_batch_factory,
        ArrowProtocolHandler,
    )
    yield session
    session.row_factory, session.client_protocol_handler = (
        row_factory,
        client_protocol_handler,
    )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cassarrow']

package_data = \
{'': ['*']}

install_requires = \
['cassandra-driver>=3.0.0',
 'pyarrow>=8.0.0',
 'pybind11>=2.10.1',
 'wheel>=0.37.0']

setup_kwargs = {
    'name': 'cassarrow',
    'version': '0.2.0rc4',
    'description': 'Apache Arrow adapter for the Cassandra python driver',
    'long_description': '[![PyPI Version][pypi-image]][pypi-url]\n[![Python Version][versions-image]][versions-url]\n[![Github Stars][stars-image]][stars-url]\n[![Build Status][build-image]][build-url]\n[![License][license-image]][license-url]\n\n\n# Cassarrow\n\nArrow based Cassandra python driver. \n\n## TLDR;\n\nSpeed up the cassandra python driver using C++ to parse cassandra queries data as [Apache Arrow](https://arrow.apache.org/) tables.\n\nKey features:\n* 20x speed up in the parsing of results\n* 14x less memory\n* Support for most native types, UDT, List and Set\n\n## Getting Started\n\n### Installation\n\n```shell\npip install cassarrow\n```\n\n### Usage\n\n```python\nimport cassarrow\nimport pyarrow as pa\n\n# ...\n\nwith cassarrow.install_cassarrow(session) as cassarrow_session:\n    table: pa.Table = cassarrow.result_set_to_table(cassarrow_session.execute("SELECT * FROM my_table"))\n```\n\n## Type Mapping\n\n### Native Types\n\n| Cassandra   | pyarrow              | Note         |\n|:------------|:---------------------|:-------------|\n| ascii       | `pa.string()`        |              |\n| bigint      | `pa.int64()`         |              |\n| blob        | `pa.binary()`        |              |\n| boolean     | `pa.bool_()`         |              |\n| counter     |                      | TODO         |\n| date        | `pa.date32()`        |              |\n| decimal     |                      | Incompatible |\n| double      | `pa.float64()`       |              |\n| duration    | `pa.duration("ns")`  |              |\n| float       | `pa.float32()`       |              |\n| inet        |                      | TODO         |\n| int         | `pa.int32()`         |              |\n| smallint    | `pa.int16()`         |              |\n| text        | `pa.string()`        |              |\n| time        | `pa.time64("ns")`    |              |\n| timestamp   | `pa.timestamp("ms")` |              |\n| timeuuid    | `pa.binary(16)`      |              |\n| tinyint     | `pa.int8()`          |              |\n| uuid        | `pa.binary(16)`      |              |\n| varchar     | `pa.string()`        |              |\n| varint      |                      | Incompatible |\n\n## Collections / UDT\n\n| Cassandra   | pyarrow     | Note   |\n|:------------|:------------|:-------|\n| list        | `pa.list_`  |        |\n| map         | `pa.map_`   |        |\n| set         | `pa.list_`  |        |\n| udt         | `pa.struct` |        |\n\n\n\n[pypi-image]: https://img.shields.io/pypi/v/cassarrow\n[pypi-url]: https://pypi.org/project/cassarrow/\n[build-image]: https://github.com/0x26res/cassarrow/actions/workflows/build.yaml/badge.svg\n[build-url]: https://github.com/0x26res/cassarrow/actions/workflows/build.yaml\n[stars-image]: https://img.shields.io/github/stars/0x26res/cassarrow\n[stars-url]: https://github.com/0x26res/cassarrow\n[versions-image]: https://img.shields.io/pypi/pyversions/cassarrow\n[versions-url]: https://pypi.org/project/cassarrow/\n[license-image]: http://img.shields.io/:license-Apache%202-blue.svg\n[license-url]: https://github.com/0x26res/cassarrow/blob/master/LICENSE.txt\n',
    'author': '0x26res',
    'author_email': '0x26res@gamil.com',
    'maintainer': '0x26res',
    'maintainer_email': '0x26res@gmail.com',
    'url': 'https://github.com/0x26res/cassarrow',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.11',
}
from build import *
build(setup_kwargs)

setup(**setup_kwargs)
